local B, L, T = unpack(select(2, ...))

T.LABEL = {
    ['bank'] = L[10001],
    ['reagent'] = L[10002],
    ['void'] = L[10003],
    ['guild'] = L[10004],
    ['player'] = L[10010],
    ['equip'] = L[10012]
}

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML bankswatch.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function BW_OnLoad(self)

        ----- REGISTER FOR EVENTS
            self:RegisterEvent('BANKFRAME_OPENED')
            self:RegisterEvent('BANKFRAME_CLOSED')
            self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
            self:RegisterEvent('VOID_STORAGE_OPEN')
            self:RegisterEvent('VOID_STORAGE_CLOSE')
            self:RegisterEvent('VOID_STORAGE_UPDATE')
            self:RegisterEvent('VOID_STORAGE_CONTENTS_UPDATE')
            self:RegisterEvent('VOID_TRANSFER_DONE')
            self:RegisterEvent('GUILDBANKFRAME_OPENED')
            self:RegisterEvent('GUILDBANKFRAME_CLOSED')
            self:RegisterEvent('GUILDBANK_UPDATE_TABS')
            self:RegisterEvent('PLAYER_ENTERING_WORLD')
            self:RegisterEvent('VARIABLES_LOADED')
            self:RegisterEvent('PLAYER_REGEN_DISABLED')
            self:RegisterEvent('GET_ITEM_INFO_RECEIVED')
            self:RegisterEvent('PLAYER_EQUIPMENT_CHANGED')

        DEFAULT_CHAT_FRAME:AddMessage(T.MSGC.DEFAULT..L[10000])

        ------ SLASH COMMANDS
            SLASH_BW1 = "/BW"
            SLASH_BW2 = "/bw"
            SLASH_BW3 = "/Bw"
            SLASH_BW4 = "/bW"
            SLASH_BW5 = "/Bankswatch"
            SLASH_BW6 = "/BanksWatch"
            SLASH_BW7 = "/BANKSWATCH"
            SLASH_BW8 = "/bankswatch"

            SlashCmdList['BW'] = B.Main

            B.CreateFrames()
            table.insert(UISpecialFrames, T.F.mainframe.bank:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.reagent:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.void:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.guild:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.search:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.character:GetName())

    end

    local bag_to_update = nil
    function BW_OnEvent(self, event, ...)
        --B.print(event)

        if event == "VARIABLES_LOADED" then
            B.print(T.MSGC.EVENT.."Variables loaded. Initializing.", 2)
            B.print(T.MSGC.INFO.."Running version: "..GetAddOnMetadata("BanksWatch", "Version"), 1)

            if BWsavedtable then
                T.BUTTON_HEIGHT = BWsavedtable['iconsize'] or 32
                T.ROW = BWsavedtable['iconrow'] or 12

                if BWsavedtable['minimize'] then T.F.mainframe.togglebutton:Show() end
                if not BWsavedtable['bankbagstoggle'] then T.F.mainframe.bank.bags:Hide() end
                if not BWsavedtable['playerbagstoggle'] then T.F.mainframe.character.bags:Hide() end
            end

            B.I.Main.Bank.Buttons()
            B.I.Main.Reagent.Buttons()
            B.I.Main.Character.Buttons()
            B.I.Main.Void.Buttons()
            B.I.Main.Guild.Buttons()

        elseif event == "GET_ITEM_INFO_RECEIVED" then
            local itemid = ...
            if T.BW.lookup[itemid] then T.BW.lookup[itemid] = nil end

            if T.BW.lookuprequest and not next(T.BW.lookup) then -- Try to minimize the amount of updates
                T.BW.lookuprequest = false
                B.BankUpdate()
                B.ReagentUpdate()
                B.VoidUpdate()
                B.GuildUpdate()
                B.PlayerUpdate()
                B.PlayerEquipmentUpdate()
            end

        elseif event == "PLAYER_ENTERING_WORLD" then
            -- When loading, BAG_UPDATE triggers, but I only want when player modifies their bank bags
            if not T.BW.loaded then
                T.BW.loaded = true
                self:RegisterEvent('PLAYER_REGEN_DISABLED')
                self:RegisterEvent('BAG_UPDATE')
                B.Initialize() -- Load or create character
                B.C.Create() -- Config
                B.SetFrameLocation()

                B.PlayerBagsScan()
                B.PlayerEquipmentScan()
                B.PlayerEquipmentUpdate()
            end

        elseif event == "PLAYERBANKSLOTS_CHANGED" then
            -- Only for generic bank bag
            local arg = ...

            if arg <= NUM_BANKGENERIC_SLOTS then -- Generic bank modified
                B.GenericBankSlotScan(arg)
            elseif arg > NUM_BANKGENERIC_SLOTS then -- Bank bag modified, will trigger BAG_UPDATE
                bag_to_update = arg - NUM_BANKGENERIC_SLOTS + 4
            end

        elseif event == "PLAYER_EQUIPMENT_CHANGED" then
            B.PlayerEquipmentSlotScan(...)

        elseif event == "BAG_UPDATE" then
            local bag = ...

            if bag == 0 or bag <= NUM_BAG_SLOTS then
                -- Only run when bags are not open.
                if not IsBagOpen(bag) then B.PlayerBagScan(bag) end
            elseif bag_to_update == bag then -- Only when actual bag is modified
                bag_to_update = nil
                B.BankBagScan(bag)
            end

        elseif event == "VOID_STORAGE_CLOSE" then
            B.VoidUpdate()

            collectgarbage()

        elseif event == "BANKFRAME_OPENED" then
            T.ISBANKOPEN = true -- Prevent BAG_UPDATE from taking effect with first login
            if not T['bank']['scanned'] then B.BankScan(); T['bank']['scanned'] = true end

        elseif event == "BANKFRAME_CLOSED" then
            T.ISBANKOPEN = false

            B.BankUpdate()
            B.ReagentUpdate()

        elseif event == "GUILDBANKFRAME_OPENED" or event == "GUILDBANK_UPDATE_TABS" then
            B.RefreshGuildTab()

        elseif event == "GUILDBANKFRAME_CLOSED" then
            B.GuildUpdate()

            collectgarbage()

        elseif event == "PLAYER_REGEN_DISABLED" then
            -- Entering combat, hide any open bags
            B.HideAll()
        end
    end

--[[//////////////////////////////////////////////////

    MAIN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Initialize()

        hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", B.ContainerFrameItemButton_OnModifiedClick)
        hooksecurefunc("ChatFrame_OnHyperlinkShow", B.ChatFrame_OnHyperlinkShow)
        hooksecurefunc("ChatEdit_InsertLink", B.InsertLink)
        hooksecurefunc("SetItemButtonQuality", B.ButtonListener)
        hooksecurefunc("CloseAllBags", B.CleanUpCrew)

        local player = UnitName('player')
        local server = GetRealmName()
        local faction = UnitFactionGroup('player')

        if (player == nil) or (player == UNKNOWNOBJECT) or (player == UNKNOWNBEING) then
            return
        end

        if BWsavedtable and BWsavedtable[server] and BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] then
            B.load(server, player, faction)
        else
            B.print(T.MSGC.EVENT.."New character detected")
            B.template(server, player, faction)
        end

        -- Reference tables for loaded character
        T['bank'] = B.GetPlayerBank('bank', true)
        T['reagent'] = B.GetPlayerBank('reagent', true)
        T['void'] = B.GetPlayerBank('void', true)
        T['guild'] = B.GetPlayerBank('guild', true)
        T['player'] = B.GetPlayerBank('player', true)
    end

    function B.Main(input)

        B.print(T.MSGC.INFO.."That's a lot of stuff.", 1)
        if input then input = string.lower(input) end

        if input == "toggle" then
            if T.F.toggleframe:IsVisible() then
                T.F.toggleframe:Hide()
            else
                T.F.toggleframe:Show()
            end

        elseif input == "debug" then
            BWdebug = not BWdebug
            B.print(T.MSGC.INFO.."Debug:|r "..T.MSGC.BOOL..tostring(BWdebug), 1)

        elseif input == "tooltip" then
            BWsavedtable.helptooltip = not BWsavedtable.helptooltip
            B.print(T.MSGC.INFO.."Help Tooltip:|r "..T.MSGC.BOOL..tostring(BWsavedtable.helptooltip), 1)
        end
    end

    function B.ButtonListener(button, quality, link)
        -- Workaround to using events which may trigger more than once.

        if B.IsGuildBank(button) then
            local page = GetCurrentGuildBankTab()
            local id = button:GetID()

            B.GuildSlotScan(page, id, button, link)
        
        elseif T.ISBANKOPEN and B.IsBankBag(button) then
            local bag = button:GetParent():GetID()
            local id = button:GetID()

            B.BankBagSlotScan(bag, id)
        
        elseif T.ISBANKOPEN and B.IsReagentBank(button) then
            local id = button:GetID()

            B.ReagentSlotScan(id)

        elseif B.IsVoidBank(button) then
            local page = tonumber(VoidStorageFrame.page)
            local id = button:GetID()

            B.VoidSlotScan(page, id, button)

        elseif B.IsPlayerBag(button) then
            local bag = button:GetParent():GetID()
            local id = button:GetID()

            B.PlayerBagSlotScan(bag, id)
        end
    end

    function B.CleanUpCrew(...)
        -- Only seems to trigger when interaction with vendor or banks is ended.
        -- perfect place to collect trash.
        B.print("Cleaning up the mess...")
        collectgarbage()
    end

    function B.ContainerFrameItemButton_OnModifiedClick(self)

        if IsShiftKeyDown() then
            local bag, item = self:GetParent():GetID(), self:GetID()
            local link = GetContainerItemLink(bag, item)

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then

                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.ChatFrame_OnHyperlinkShow(self, link, text, button)

        if IsShiftKeyDown() and link then
            if string.find(link, 'achievement:*') or string.find(link,'spell:*') then
                return false
            end

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.InsertLink(text)

        if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
            T.F.mainframe.searchbar:SetText(string.find(text, "^|c%x+|H(.+)|h%[.*%]"))
        end
    end

    function B.dropDownMenuItem_OnClick(self)
        local server, player, faction
        local value = self.value

        if value == "delete" then
            BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)] = nil
        else
            server, player = value:match"([^.]*).(.*)"
            player, faction = player:match"([^.]*).(.*)"
            faction = faction == "H" and "Horde" or faction == "A" and "Alliance" or faction == "N" and "Neutral"
            B.load(server, player, faction)
        end

        BWdropDownMenuButton:Click()
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS (OLDER)

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.ReagentScan()

        local x, page, itemID
        local reagent = T['reagent']

        for page = 1, T.REAGENT_PAGES do
            if not reagent[page] then reagent[page] = B.newtable() end

            for x = 1, T.NUM_REAGENT_SLOTS do

                local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

                if link then
                    reagent[page][x] = reagent[page][x] or B.newtable()
                    reagent[page][x]['link'] = link
                    reagent[page][x]['count'] = count > 1 and count or nil
                else
                    reagent[page][x] = B.deltable(reagent[page][x])
                end
            end
        end
    end

    function B.VoidScan()

        if not IsVoidStorageReady() then return end
        local x, page, itemID
        local void = T['void']

        for page = 1, T.VOID_PAGES do
            if not void[page] then void[page] = B.newtable() end

            for x = 1, T.NUM_VOID_SLOTS do
                itemID = GetVoidItemInfo(page, x)

                if itemID then
                    void[page][x] = void[page][x] or B.newtable()
                    void[page][x]['link'] = page == 1 and GetVoidItemHyperlinkString(x) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + x)
                else
                    void[page][x] = B.deltable(void[page][x])
                end
            end
        end
    end

    function B.GuildScan()

        local x, i, id, count, item, button
        local guild = T['guild']
        local page = GetCurrentGuildBankTab()

        if guild[page].access then
            for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                _, count = GetGuildBankItemInfo(page, x)
                if count > 0 then
                    guild[page][x] = guild[page][x] or B.newtable()
                    guild[page][x]['link'] = GetGuildBankItemLink(page, x)
                    guild[page][x]['count'] = count > 1 and count or nil
                else
                    guild[page][x] = B.deltable(guild[page][x])
                end
            end
        end
    end

    function B.PlayerBagsScan()

        local x, counter, count, link, bag, slot, slots
        local player = T['player']

        for bag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do
            if not player[bag] then player[bag] = B.newtable() end

            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    player[bag][x] = player[bag][x] or B.newtable()
                    player[bag][x]['count'] = count
                    player[bag][x]['link'] = link
                else
                    player[bag][x] = B.deltable(player[bag][x])
                end
            end
        end
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.BankScan()

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, counter, count, link, bag, slot, slots
        local bank = T['bank']

        if not bank[1] then bank[1] = B.newtable() end
        for x = 1, NUM_BANKGENERIC_SLOTS do
            _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, x)

            if link then
                bank[1][x] = bank[1][x] or B.newtable()
                bank[1][x]['count'] = count
                bank[1][x]['link'] = link
            else
                bank[1][x] = B.deltable(bank[1][x])
            end
        end

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            if not bank[bag] then bank[bag] = B.newtable() end
            
            local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
            local baglink = GetInventoryItemLink("player", inventoryID)

            slots = GetContainerNumSlots(bag)
            bank[bag]['slots'] = slots
            bank[bag]['link'] = baglink

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    bank[bag][x] = bank[bag][x] or B.newtable()
                    bank[bag][x]['link'] = link
                    bank[bag][x]['count'] = count > 1 and count or nil
                else
                    bank[bag][x] = B.deltable(bank[bag][x])
                end
            end
        end
    end

    function B.BankBagScan(bag)

        local x, slots, count, link
        local bank = T['bank']

        if not bank[bag] then bank[bag] = B.newtable() end
            
        local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
        local baglink = GetInventoryItemLink("player", inventoryID)

        slots = GetContainerNumSlots(bag)
        bank[bag]['slots'] = slots
        bank[bag]['link'] = baglink

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

            if link then
                bank[bag][x] = bank[bag][x] or B.newtable()
                bank[bag][x]['link'] = link
                bank[bag][x]['count'] = count > 1 and count or nil
            else
                bank[bag][x] = B.deltable(bank[bag][x])
            end
        end
    end

    function B.GenericBankSlotScan(slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, slot)

        if link then
            bank[1][slot] = bank[1][slot] or B.newtable()
            bank[1][slot]['link'] = link
            bank[1][slot]['count'] = count > 1 and count or nil
        else
            bank[1][slot] = B.deltable(bank[1][slot])
        end
    end

    function B.BankBagSlotScan(bag, slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        if not bank[bag] then bank[bag] = B.newtable() end

        if link then
            bank[bag][slot] = bank[bag][slot] or B.newtable()
            bank[bag][slot]['link'] = link
            bank[bag][slot]['count'] = count > 1 and count or nil
        else
            bank[bag][slot] = B.deltable(bank[bag][slot])
        end
    end

    function B.ReagentSlotScan(slot)

        local reagent = T['reagent']
        local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

        if not reagent[1] then reagent[1] = B.newtable() end

        if link then
            reagent[1][slot] = reagent[1][slot] or B.newtable()
            reagent[1][slot]['link'] = link
            reagent[1][slot]['count'] = count > 1 and count or nil
        else
            reagent[1][slot] = B.deltable(reagent[1][slot])
        end
    end

    function B.VoidSlotScan(page, slot, button)

        local void = T['void']

        if not void[page] then void[page] = B.newtable() end

        if button.hasItem then
            void[page][slot] = void[page][slot] or B.newtable()
            void[page][slot]['link'] = page == 1 and GetVoidItemHyperlinkString(slot) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + slot)
        else
            void[page][slot] = B.deltable(void[page][slot])
        end
    end

    function B.GuildSlotScan(page, slot, button, link)

        local guild = T['guild']

        if button.count > 0 then
                
            if strfind(link, "item:82800:") then -- Caged pet
                link = B.GetGuildPetLink(page, slot)
            end

            guild[page][slot] = guild[page][slot] or B.newtable()
            guild[page][slot]['count'] = button.count > 1 and button.count or nil
            guild[page][slot]['link'] = link
        else
            guild[page][slot] = B.deltable(guild[page][slot])
        end
    end

    function B.PlayerBagScan(bag)

        local x, slots, count, link
        local player = T['player']

        if not player[bag] then player[bag] = B.newtable() end
            
        slots = GetContainerNumSlots(bag)
        player[bag]['slots'] = slots

        if bag > 0 then
            local inventoryID = ContainerIDToInventoryID(bag)
            local baglink = GetInventoryItemLink("player", inventoryID)
            player[bag]['link'] = baglink
        end

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

            if link then
                player[bag][x] = player[bag][x] or B.newtable()
                player[bag][x]['count'] = count
                player[bag][x]['link'] = link
            else
                player[bag][x] = B.deltable(player[bag][x])
            end
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
    end

    function B.PlayerBagSlotScan(bag, slot)

        local player = T['player']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end

        if link then
            player[bag][slot] = player[bag][slot] or B.newtable()
            player[bag][slot]['link'] = link
            player[bag][slot]['count'] = count > 1 and count or nil
        else
            player[bag][slot] = B.deltable(player[bag][slot])
        end

        if B.IsCurrentPlayer() then B.PlayerSlotUpdate(bag, slot) end
    end

    function B.PlayerEquipmentScan()

        local inventory, slotId, link
        local player = T['player']

        if not player[-1] then player[-1] = B.newtable() end

        inventory = {'HeadSlot', 'NeckSlot', 'ShoulderSlot', 'BackSlot', 'ChestSlot', 'ShirtSlot',
                     'TabardSlot', 'WristSlot', 'HandsSlot', 'WaistSlot', 'LegsSlot', 'FeetSlot',
                     'Finger0Slot', 'Finger1Slot', 'Trinket0Slot', 'Trinket1Slot', 'MainHandSlot',
                     'SecondaryHandSlot'}

        -- Tabard slot is slotId 19?
        for x = 1, #inventory do

            slotId = GetInventorySlotInfo(inventory[x])
            link = GetInventoryItemLink("player", slotId)

            if slotId == 19 then slotId = 18 end

            if link then
                player[-1][slotId] = player[-1][slotId] or B.newtable()
                player[-1][slotId]['link'] = link
            else
                player[-1][slotId] = B.deltable(player[-1][slotId])
            end
        end
    end

    function B.PlayerEquipmentSlotScan(slot)

        local player = T['player']
        local link = GetInventoryItemLink("player", slot)

        if slot == 19 then slot = 18 end

        if link then
            player[-1][slot] = player[-1][slot] or B.newtable()
            player[-1][slot]['link'] = link
        else
            player[-1][slot] = B.deltable(player[-1][slot])
        end

        if B.IsCurrentPlayer() then B.PlayerEquipmentSlotUpdate(slot) end
    end

--[[//////////////////////////////////////////////////

    UPDATE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.ScrollbarUpdate()
        local r, g, b, idx, link, hexcolor, itemRarity
        local total_items = #T.BW.items
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.search.scroll)

        FauxScrollFrame_Update(T.F.mainframe.search.scroll, total_items, B.RowsToFit(), T.BUTTON_HEIGHT)
        
        for x = 1, B.RowsToFit() do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.search.list.buttons[x])
            T.F.mainframe.search.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.search.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.search.list.buttons[x]:Hide()
            T.F.mainframe.search.enddot:Hide()
        end

        for x = 1, B.RowsToFit() do
            -- Get the appropriate item, which will be x + value
            if (T.F.mainframe.layout == 2 or T.F.mainframe.layout == 4) and total_items > 0 and total_items < B.RowsToFit() then
                local difference = B.RowsToFit() - total_items
                buttonnum = x + difference
            else
                buttonnum = x
            end

            idx = x + currentscrollbarvalue

            if T.BW.items[idx] then
                r, g, b, hexcolor = GetItemQualityColor(T.BW.items[idx].quality)

                T.F.mainframe.search.list.buttons[buttonnum].quest = T.BW.items[idx].quest
                T.F.mainframe.search.list.buttons[buttonnum].quality = T.BW.items[idx].quality
                T.F.mainframe.search.list.buttons[buttonnum].icon:SetTexture(T.BW.items[idx].icon)
                T.F.mainframe.search.list.buttons[buttonnum].leftstring:SetText("|c"..hexcolor..T.BW.items[idx].name)

                if T.BW.items[idx].count > 1 then
                    T.F.mainframe.search.list.buttons[buttonnum].text:SetText(T.BW.items[idx].count)
                end

                if T.BW.items[idx].ilvl and T.BW.items[idx].ilvl ~= "" then
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetText(T.BW.items[idx].ilvl)
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetTextColor(r, g, b, 1)
                end

                --loc, num = T.BW.items[idx].loc:match"([^.]*).(.*)"
                if T.BW.items[idx].loc == "alt" then
                    location = T.BW.items[idx].num
                else
                    location = T.BW.items[idx].num == 1 and T.LABEL[T.BW.items[idx].loc] or T.LABEL[T.BW.items[idx].loc].." "..T.BW.items[idx].num
                end
                T.F.mainframe.search.list.buttons[buttonnum].location = {['loc'] = T.BW.items[idx].loc, ['num'] = T.BW.items[idx].num}
                T.F.mainframe.search.list.buttons[buttonnum].rightstring:SetText(location)
                T.F.mainframe.search.list.buttons[buttonnum].link = T.BW.items[idx].link
                T.F.mainframe.search.list.buttons[buttonnum]:Show()
                T.F.mainframe.search.enddot:Show()
                if idx == total_items then
                    T.F.mainframe.search.enddot:Hide()
                end
            end
        end
    end

    function B.BankUpdate()

        local page, total, counter, quality, r, g, b
        local bank = B.GetPlayerBank('bank')

        total = 0
        for page = 1, #bank do
            total = total + (bank[page] and bank[page]['slots'] or 0)
        end
        B.ValidateBags(total, T.F.mainframe.bank, B.I.bank)

        counter = 0
        for page = 1, #bank do
            if bank[page] then

                if page > 1 then -- Ignore generic bank
                    B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, bank[page])
                end
            
                B.UpdatePage(T.F.mainframe.bank, bank[page]['slots'], bank[page], counter)
                counter = counter + bank[page]['slots']
            end
        end
    end

    function B.PlayerUpdate()

        local page, total, counter, quality, r, g, b
        local player = B.GetPlayerBank('player')

        total = 0
        for page = 0, #player do
            total = total + (player[page] and player[page]['slots'] or 0)

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end
        end
        B.ValidateBags(total, T.F.mainframe.character, B.I.character)

        counter = 0
        for page = 0, #player do
            if player[page] then

                B.UpdatePage(T.F.mainframe.character, player[page]['slots'], player[page], counter)
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.PlayerSlotUpdate(bag, slot)

        local page, counter, item
        local player = B.GetPlayerBank('player')

        total = 0
        for page = 0, #player do
            total = total + (player[page] and player[page]['slots'] or 0)

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end
        end
        B.ValidateBags(total, T.F.mainframe.character, B.I.character)

        counter = 0
        for page = 0, #player do
            if player[page] then
                if page == bag then
                    
                    B.ResetButton(T.F.mainframe.character.buttons[counter + slot])
                    if player[bag][slot] then

                        item = B.GetItem(player[bag][slot].link)
                        if next(item) then B.UpdateButton(T.F.mainframe.character.buttons[counter + slot], item, player[bag][slot]) end
                    end
                end
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.PlayerEquipmentUpdate()
        local player = B.GetPlayerBank('player')

        B.UpdatePage(T.F.mainframe.character.equipment, 18, player[-1])
    end

    function B.PlayerEquipmentSlotUpdate(slot)
        local player = B.GetPlayerBank('player')

        B.ResetButton(T.F.mainframe.character.equipment.buttons[slot])
        if player[-1][slot] then

            item = B.GetItem(player[-1][slot].link)
            if next(item) then B.UpdateButton(T.F.mainframe.character.equipment.buttons[slot], item, player[-1][slot]) end
        end
    end

    function B.ReagentUpdate()

        local page
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            B.UpdatePage(T.F.mainframe.reagent[page], T.NUM_REAGENT_SLOTS, reagent[page])
        end
    end

    function B.VoidUpdate()

        local page
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do
            B.UpdatePage(T.F.mainframe.void[page], T.NUM_VOID_SLOTS, void[page])
        end
    end

    function B.GuildUpdate()

        local page
        local guild = B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do
            if guild[page] then

                if guild[page].icon then
                    T.F.mainframe.guild[page].tab:SetNormalTexture(guild[page].icon)
                    T.F.mainframe.guild[page].tab.name = guild[page].name
                    T.F.mainframe.guild[page].tab:Show()
                else
                    T.F.mainframe.guild[page].tab:Hide()
                end

                if guild[page].access then -- Validate user has access to view tab
                    T.F.mainframe.guild[page].tab:Enable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(false)
                    B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, guild[page])
                else
                    T.F.mainframe.guild[page].tab:Disable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(true)
                end
            else
                T.F.mainframe.guild[page].tab:Hide()
            end
        end
    end

    function B.RefreshGuildTab()

        local x, name, icon, isViewable
        local guild = T['guild']
        local tabs = GetNumGuildBankTabs()

        if IsInGuild() then guild['name'] = GetGuildInfo('player') end

        for x = 1, MAX_GUILDBANK_TABS do
            if x <= tabs then
                if not guild[x] then guild[x] = B.newtable() end
                name, icon, isViewable = GetGuildBankTabInfo(x)
                guild[x]['icon'] = icon
                guild[x]['name'] = name
                guild[x]['access'] = isViewable

                T.F.mainframe.guild[x].tab:SetNormalTexture(icon)
                T.F.mainframe.guild[x].tab.name = name
                T.F.mainframe.guild[x].tab:Show()
            else
                guild[x] = B.deltable(guild[x])
                T.F.mainframe.guild[x].tab:Hide()
            end
        end
    end

    function B.UpdateBagTab(tab, data)

        if data.link then
            quality = B.GetQualityFromLink(data.link)
            r, g, b = GetItemQualityColor(quality)

            tab:SetBackdropBorderColor(r, g, b, 1)
            tab.link = data.link
            tab.quality = quality
            tab:SetNormalTexture(GetItemIcon(data.link))
            tab:Show()
        else
            tab:Hide()
        end
    end

    function B.UpdatePage(page, maxslots, data, optstart)
        local x, item
        optstart = optstart or 0

        for x = 1, maxslots do

            B.ResetButton(page.buttons[x + optstart])
            if data and data[x] then

                item = B.GetItem(data[x].link)
                if next(item) then B.UpdateButton(page.buttons[x + optstart], item, data[x]) end
            end
        end
    end

    function B.GetItem(itemlink)

        item = strmatch(itemlink, "|Hbattlepet:") and B.GetPetInfo(itemlink) or {GetItemInfo(itemlink)}

        if not next(item) then -- Item is not cached yet
            T.BW.lookuprequest = true
            T.BW.lookup[tonumber(B.GetItemIdFromLink(itemlink))] = true
        end

        return item
    end

    function B.UpdateButton(button, item, data)
        local r, g, b, ilvl

        r, g, b = GetItemQualityColor(item[3])

        button.name = item[1]
        button.count = data.count or 1
        button.quality = item[3]
        button.link = item[2]
        button.class = item[6]
        button.subclass = item[7]
        button.equipslot = item[9]

        button.icon:SetTexture(item[10])
        button.icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        
        if (data.count or 1) > 1 then button.text:SetText(data.count) end
        if item[3] > 1 then button:SetBackdropBorderColor(r, g, b, 1) end

        if item[6] == BATTLE_PET_SOURCE_2 then -- QUEST
            button:SetBackdropBorderColor(1, 1, 0, 1)
            button.quest = true

        elseif item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON then -- ILVL
            ilvl = B.DetectILVL(item[2])
            if ilvl then
                button.ilvl:SetText(ilvl)
                button.ilvl:SetTextColor(r, g, b, 1)
            else
                button.ilvl:SetText(item[4])
                button.ilvl:SetTextColor(r, g, b, 1)
            end
        end
    end

    function B.ResetButton(button)

        button.icon:SetTexture("")
        button:SetBackdropBorderColor(0, 0, 0, 1)
        button.link = nil
        button.quality = nil
        button.quest = nil
        button.name = nil
        button.count = nil
        button.text:SetText("")
        button.ilvl:SetText("")
    end

--[[//////////////////////////////////////////////////

    SEARCH FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Search(text)

        if text == "" or text == SEARCH then text = false end

        T.F.mainframe.search.scroll.ScrollBar:SetValue(0)

        T.BW.items = B.deltable(T.BW.items)
        T.BW.items = B.newsearchtable()

        B.BankSearch(text)
        B.ReagentSearch(text)
        B.VoidSearch(text)
        B.GuildSearch(text)
        B.PlayerSearch(text)
        B.PlayerEquipmentSearch(text)
        B.SearchAlts(text)

        B.ScrollbarUpdate()

        if text and T.F.mainframe.countbutton.count > 0 then
            T.F.mainframe.countbutton.labelcount:SetText("|cffffffff"..T.F.mainframe.countbutton.count.."|r")
            T.F.mainframe.countbutton:Show()
            if not T.F.mainframe.search:IsVisible() and not T.F.mainframe.lastbank then
                T.F.mainframe.countbutton:Click()
            end
        else
            T.F.mainframe.countbutton:Hide()
        end
    end

    function B.BankSearch(text)

        local x, page, counter, button, buttons, name
        local bank = B.GetPlayerBank('bank')

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if page > 1 then -- ignore generic bank
                    local tab = page - 4

                    if text then T.F.mainframe.bank.bags[tab].tab.search:Show()
                    else T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.bank.buttons

                for x = 1, bank[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "bank."..(page == 1 and page or page - 4)
                            B.SearchSlot(button)
                            if page > 1 then local tab = page - 4; T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerSearch(text)

        local x, page, counter, button, buttons, name
        local player = B.GetPlayerBank('player')

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    local tab = page

                    if text then T.F.mainframe.character.bags[tab].tab.search:Show()
                    else T.F.mainframe.character.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "player."..(page == 0 and 1 or page + 1)
                            B.SearchSlot(button)
                            if page > 0 then local tab = page; T.F.mainframe.character.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerEquipmentSearch(text)

        local x, button, buttons, name
        local player = B.GetPlayerBank('player')

        if not player[-1] then return end

        buttons = T.F.mainframe.character.equipment.buttons
        for x = 1, #player[-1] do
            button = buttons[x]

            if button.name and text then
                if B.find(text, button) then
                    button.loc = "equip.1"
                    B.SearchSlot(button)
                else
                    button.search:Show()
                end
            else
                button.search:Hide()
            end

        end
    end

    function B.BankBagSearch(bag)

        local bank = B.GetPlayerBank('bank')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if not bag or page == bag then
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.PlayerBagSearch(bag)

        local player = B.GetPlayerBank('player')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if not bag or page == bag then
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.ReagentSearch(text)

        local x, page, button, buttons
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            buttons = T.F.mainframe.reagent[page].buttons
            for x = 1, T.NUM_REAGENT_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "reagent."..page
                        B.SearchSlot(button)
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.VoidSearch(text)

        local x, page, button, buttons
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do

            if text then T.F.mainframe.void[page].tab.search:Show() 
            else T.F.mainframe.void[page].tab.search:Hide() end

            buttons = T.F.mainframe.void[page].buttons

            for x = 1, T.NUM_VOID_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "void."..page
                        B.SearchSlot(button)
                        T.F.mainframe.void[page].tab.search:Hide()
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.GuildSearch(text)

        local x, page, button, buttons
        local guild = B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do

            if guild[page] and guild[page].access then

                if text then T.F.mainframe.guild[page].tab.search:Show()
                else T.F.mainframe.guild[page].tab.search:Hide() end

                buttons = T.F.mainframe.guild[page].buttons

                for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do
                    button = buttons[x]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "guild."..page
                            B.SearchSlot(button)
                            T.F.mainframe.guild[page].tab.search:Hide()
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                end
            end
        end
    end

    function B.SearchAlts(text)
        local player, faction, loc, num, name, link, item
        -- Only same server and faction
        if not text then return end

        if BWsavedtable then
            for k, v in pairs(BWsavedtable[T.ACTIVE_TABLE.server]) do
                player, faction = k:match"([^.]*).(.*)"
                if faction == string.sub(T.ACTIVE_TABLE.faction, 1, 1) and player ~= T.ACTIVE_TABLE.player then
                    for i, j in pairs(v) do-- banks, bags
                        for m, n in pairs(j) do -- bags, items
                            if type(n) == "table" then
                                for x = 1, table.maxn(n) do
                                    if n[x] and next(n[x]) then

                                        link = n[x]['link']
                                        item = strmatch(link, "|Hbattlepet:") and B.GetPetInfo(link) or {GetItemInfo(link)}
                                        temp = B.newtable()
                                        temp.name = item[1]
                                        temp.quality = item[3]
                                        temp.link = item[2]
                                        temp.class = item[6]
                                        temp.subclass = item[7]
                                        temp.equipslot = item[9]

                                        if B.find(text, temp) then
                                            B.SearchAltSlot(player, item, n[x]['count'] or 1)
                                        end

                                        B.deltable(temp)
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end
    end

    function B.SearchSlot(button)

        button.search:Hide()

        loc, num = button.loc:match"([^.]*).(.*)"
        num = tonumber(num)

        result = T.BW.items(loc, num, button.name) -- use metatable __call to group items
        if not next(result) then -- New entry
            setmetatable(result, {__index = {
                ['loc'] = loc,
                ['num'] = num,
                ['name'] = button.name,
                ['link'] = button.link,
                ['quest'] = button.quest,
                ['quality'] = button.quality,
                ['ilvl'] = button.ilvl:GetText(),
                ['icon'] = button.icon:GetTexture()}})

            result['count'] = 0
        end

        result['count'] = result['count'] + button.count
        T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + button.count
    end

    function B.SearchAltSlot(alt, item, count)

        result = T.BW.items("alt", alt, item[1]) -- use metatable __call to group items
        if not next(result) then -- New entry
            setmetatable(result, {__index = {
                ['loc'] = "alt",
                ['num'] = alt,
                ['name'] = item[1],
                ['link'] = item[2],
                ['quality'] = item[3],
                ['quest'] = item[6] == BATTLE_PET_SOURCE_2,
                ['ilvl'] = item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON and item[4] or nil,
                ['icon'] = item[10]}})

            result['count'] = 0
        end

        result['count'] = result['count'] + count
    end
